/*************************************************************************
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 2007      Bryan Christ <bryan.christ@hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/


#include <stdio.h>
#include <string.h>

#include "psthread.h"
#include "pseudo.h"

PSTHREAD_Q* psthread_q_init(const gchar *queue_name)
{
	static GStaticRecMutex	queue_list_lock=G_STATIC_REC_MUTEX_INIT;
	static GSList			*queue_list=NULL;
	static guint8			queue_list_length=0;
	GSList					*node=NULL;
	PSTHREAD_Q				*psthread_q;
	gint					length;
		
	/* lock the master list of queues */
	g_static_rec_mutex_lock(&queue_list_lock);
	
	/* set a default queue */
	if(queue_name!=NULL) queue_name="default";
	
	/* create a new ps_thread_q if the master list of queues is empty */
	if(queue_list_length==0)
	{
		queue_list=NULL;
		psthread_q=(PSTHREAD_Q*)g_malloc0(sizeof(PSTHREAD_Q));
		queue_list=g_slist_append(queue_list,(gpointer)psthread_q);
		g_static_rec_mutex_init(&psthread_q->queue_lock);
		psthread_q->queue_name=queue_name;
		queue_list_length++;		
		
		g_static_rec_mutex_unlock(&queue_list_lock);
		return psthread_q;
	}
	
	/* look for an already existing queue */	
	node=queue_list;
	length=strlen(queue_name);
	while(node!=NULL)
	{
		psthread_q=(PSTHREAD_Q*)node->data;
		if(memcmp(queue_name,psthread_q->queue_name,length)==0) break;
		node=node->next;
	}
	
	/* return an existing queue */
	if(node!=NULL)
	{
		g_static_rec_mutex_unlock(&queue_list_lock);
		return psthread_q;
	}
	
	/* create and add a new queue */	
	psthread_q=(PSTHREAD_Q*)g_malloc0(sizeof(PSTHREAD_Q));
	queue_list=g_slist_prepend(queue_list,(gpointer)psthread_q);
	psthread_q->queue_name=queue_name;
	g_static_rec_mutex_init(&psthread_q->queue_lock);
	queue_list_length++;

	/* unlock master list and return */	
	g_static_rec_mutex_unlock(&queue_list_lock);	
	return psthread_q;
}

void psthread_run(const gchar *queue_name,guint32 count)
{
	PSTHREAD_Q		*psthread_q=NULL;
	PSTHREAD_UNION	*psthread_union=NULL;
	GSList			*node;
	gint			result;
	gint			priority;
	gint			i;
	
	psthread_q=psthread_q_get(queue_name);
	if(psthread_q==NULL) return;
	
	g_static_rec_mutex_lock(&psthread_q->queue_lock);
	if(count==0) count=g_slist_length(psthread_q->queue);
	if(count==0)
	{
		g_static_rec_mutex_unlock(&psthread_q->queue_lock);
		return;
	}
		
	node=psthread_q->queue;
	while(count!=0 && node!=NULL)
	{
		psthread_union=(PSTHREAD_UNION*)node->data;
		psthread_q->queue=g_slist_delete_link(psthread_q->queue,node);
		priority=psthread_union->priority;
		
		for(i=0;i<priority;i++)
		{
			result=psthread_union->psthread(psthread_union->arg1,
				psthread_union->arg2);
			if(result==PSTHREAD_TERMINATE)
			{
				psthread_q->vcount-=(psthread_union->priority-1);
				g_free(psthread_union);
				break;
			}
			if(result==PSTHREAD_GROW && psthread_union->priority<4)
			{
				psthread_union->priority++;
				psthread_q->vcount++;
				continue;
			}
			if(result==PSTHREAD_CONTINUE)
			{
				psthread_union->priority--;
				if(psthread_union->priority<1) psthread_union->priority=1;
				else psthread_q->vcount--;
				continue;
			}
		}
				
		if(result!=PSTHREAD_TERMINATE)
			psthread_q->queue=g_slist_append(psthread_q->queue,
				(gpointer)psthread_union);
			
		node=psthread_q->queue;
		count--;
	}
	
	g_static_rec_mutex_unlock(&psthread_q->queue_lock);		
	return;
}
	
guint32 psthread_add(const gchar *queue_name,PSTHREAD psthread,
	gpointer arg1,gpointer arg2)
{
	PSTHREAD_Q		*psthread_q=NULL;
	PSTHREAD_UNION	*psthread_union=NULL;
	
	psthread_q=psthread_q_get(queue_name);
	if(psthread_q==NULL) return 0;
	if(psthread==NULL) return 0;
	
	psthread_union=(PSTHREAD_UNION*)g_malloc0(sizeof(PSTHREAD_UNION));
	psthread_union->psthread=psthread;
	psthread_union->arg1=arg1;
	psthread_union->arg2=arg2;
	psthread_union->uid=(guint32)psthread_union;
	psthread_union->priority=1;
	
	g_static_rec_mutex_lock(&psthread_q->queue_lock);
	psthread_q->queue=g_slist_append(psthread_q->queue,
		(gpointer)psthread_union);
	g_static_rec_mutex_unlock(&psthread_q->queue_lock);
	
	return psthread_union->uid;
}

void	 psthread_del(const gchar *queue_name,guint32 uid)
{
	PSTHREAD_Q		*psthread_q=NULL;
	PSTHREAD_UNION	*psthread_union;
	GSList			*node;
		
	psthread_q=psthread_q_get(queue_name);
	if(psthread_q==NULL) return;
	if(uid==0) return;
	
	g_static_rec_mutex_lock(&psthread_q->queue_lock);
	node=psthread_q->queue;
	while(node!=NULL)
	{
		psthread_union=(PSTHREAD_UNION*)node->data;
		if(psthread_union->uid==uid) break;
		node=node->next;
	}
		
	if(node!=NULL)
	{
		psthread_q->vcount-=(psthread_union->priority-1);
		psthread_q->queue=g_slist_delete_link(psthread_q->queue,node);
		g_free(psthread_union);
	}
		
	g_static_rec_mutex_unlock(&psthread_q->queue_lock);
	return;
}

gint psthread_count(const gchar *queue_name,guint *pcount,guint *vcount)
{
	PSTHREAD_Q		*psthread_q=NULL;
		
	psthread_q=psthread_q_get(queue_name);
	if(psthread_q==NULL) return -1;
	if(pcount==NULL && vcount==NULL) return -1;

	g_static_rec_mutex_lock(&psthread_q->queue_lock);
	if(pcount!=NULL) *pcount=g_slist_length(psthread_q->queue);
	if(vcount!=NULL) *vcount=psthread_q->vcount;
	g_static_rec_mutex_unlock(&psthread_q->queue_lock);

	return 0;
}
